@@ -22,6 +22,7 @@
 #include <config.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <string.h>
 #include <assert.h>
 
 #include "util.h"
@@ -458,83 +459,124 @@ parse_user_id ( struct user_id_s *uid, char *tail )
 
 }
 
+static void
+parse_x509_user_id ( struct user_id_s *uid, char *tail )
+{
+  const char *s;
+
+  s=uid->name; 
+  if (*s == '<' && s[strlen(s)-1] == '>')
+    uid->email_part = s;
+  
+  /* let unused parts point to an EOS */ 
+  tail--;
+  if (!uid->name_part)
+    uid->name_part = tail;
+  if (!uid->email_part)
+    uid->email_part = tail;
+  if (!uid->comment_part)
+    uid->comment_part = tail;
+}
+
 /* 
  * Take a name from the --with-colon listing, remove certain escape sequences
  * sequences and put it into the list of UIDs
  */
 GpgmeError
-_gpgme_key_append_name ( GpgmeKey key, const char *s )
+_gpgme_key_append_name (GpgmeKey key, const char *s)
 {
-    struct user_id_s *uid;
-    char *d;
+  struct user_id_s *uid;
+  char *d;
 
-    assert (key);
-    /* we can malloc a buffer of the same length, because the
-     * converted string will never be larger. Actually we allocate it
-     * twice the size, so that we are able to store the parsed stuff
-     * there too */
-    uid = xtrymalloc ( sizeof *uid + 2*strlen (s)+3 );
-    if ( !uid )
-        return mk_error (Out_Of_Core);
-    uid->revoked = 0;
-    uid->invalid = 0;
-    uid->validity = 0;
-    uid->name_part = NULL;
-    uid->email_part = NULL;
-    uid->comment_part = NULL;
-    d = uid->name;
+  assert (key);
+  /* We can malloc a buffer of the same length, because the converted
+     string will never be larger. Actually we allocate it twice the
+     size, so that we are able to store the parsed stuff there too.  */
+  uid = xtrymalloc ( sizeof *uid + 2*strlen (s)+3);
+  if (!uid)
+    return mk_error (Out_Of_Core);
+  uid->revoked = 0;
+  uid->invalid = 0;
+  uid->validity = 0;
+  uid->name_part = NULL;
+  uid->email_part = NULL;
+  uid->comment_part = NULL;
+  uid->next = NULL;
+  d = uid->name;
 
-    while ( *s ) {
-        if ( *s != '\\' )
-            *d++ = *s++;
-        else if ( s[1] == '\\' ) {
-            s++;
-            *d++ = *s++; 
+  while (*s)
+    {
+      if (*s != '\\')
+	*d++ = *s++;
+      else if (s[1] == '\\')
+	{
+	  s++;
+	  *d++ = *s++; 
         }
-        else if ( s[1] == 'n' ) {
-            s += 2;
-            *d++ = '\n'; 
+      else if (s[1] == 'n')
+	{
+	  s += 2;
+	  *d++ = '\n'; 
         }
-        else if ( s[1] == 'r' ) {
-            s += 2;
-            *d++ = '\r'; 
+      else if (s[1] == 'r')
+	{
+	  s += 2;
+	  *d++ = '\r'; 
         }
-        else if ( s[1] == 'v' ) {
-            s += 2;
-            *d++ = '\v'; 
+      else if (s[1] == 'v')
+	{
+	  s += 2;
+	  *d++ = '\v'; 
         }
-        else if ( s[1] == 'b' ) {
-            s += 2;
-            *d++ = '\b'; 
+      else if (s[1] == 'b')
+	{
+	  s += 2;
+	  *d++ = '\b'; 
         }
-        else if ( s[1] == '0' ) {
-            /* Hmmm: no way to express this */
-            s += 2;
-            *d++ = '\\';
-            *d++ = '\0'; 
+      else if (s[1] == '0')
+	{
+	  /* Hmmm: no way to express this */
+	  s += 2;
+	  *d++ = '\\';
+	  *d++ = '\0'; 
         }
-        else if ( s[1] == 'x' && my_isdigit (s[2]) && my_isdigit (s[3]) ) {
-            unsigned int val = (s[2]-'0')*16 + (s[3]-'0');
-            if ( !val ) {
-                *d++ = '\\';
-                *d++ = '\0'; 
+      else if (s[1] == 'x' && my_isdigit (s[2]) && my_isdigit (s[3]))
+	{
+	  unsigned int val = (s[2]-'0')*16 + (s[3]-'0');
+	  if (!val)
+	    {
+	      *d++ = '\\';
+	      *d++ = '\0'; 
             }
-            else 
-                *(byte*)d++ = val;
-            s += 3;
+	  else 
+	    *(byte*)d++ = val;
+	  s += 3;
         }
-        else { /* should not happen */
-            s++;
-            *d++ = '\\'; 
-            *d++ = *s++;
+      else
+	{
+	  /* should not happen */
+	  s++;
+	  *d++ = '\\'; 
+	  *d++ = *s++;
         } 
     }
-    *d++ = 0;
-    parse_user_id ( uid, d );
+  *d++ = 0;
+  if (key->x509)
+    parse_x509_user_id (uid, d);
+  else
+    parse_user_id (uid, d);
 
-    uid->next = key->uids;
+  if (key->uids)
+    {
+      struct user_id_s *u = key->uids;
+      while (u->next)
+	u = u->next;
+      u->next = uid;
+    }
+  else
     key->uids = uid;
-    return 0;
+
+  return 0;
 }
 
 
@@ -641,19 +683,12 @@ gpgme_key_get_as_xml ( GpgmeKey key )
     add_tag_and_uint (d, "algo", key->keys.key_algo );
     add_tag_and_uint (d, "len", key->keys.key_len );
     add_tag_and_time (d, "created", key->keys.timestamp );
-    /*add_tag_and_time (d, "expires", key->expires );*/
+    add_tag_and_time (d, "expire", key->keys.expires_at );
     _gpgme_data_append_string (d, "  </mainkey>\n");
 
-    /* Now the user IDs.  We are listing the last one first because this is
-     * the primary one. */
-    for (u = key->uids; u && u->next; u = u->next )
-        ;
-    if (u) {
-        one_uid_as_xml (d,u);
-        for ( u = key->uids; u && u->next; u = u->next ) {
-            one_uid_as_xml (d,u);
-        }
-    }
+    /* Now the user IDs.  */
+    for (u = key->uids; u; u = u->next)
+      one_uid_as_xml (d,u);
 
     /* and now the subkeys */
     for (k=key->keys.next; k; k = k->next ) {
@@ -674,6 +709,7 @@ gpgme_key_get_as_xml ( GpgmeKey key )
         add_tag_and_uint (d, "algo", k->key_algo );
         add_tag_and_uint (d, "len", k->key_len );
         add_tag_and_time (d, "created", k->timestamp );
+        add_tag_and_time (d, "expire", k->expires_at );
         _gpgme_data_append_string (d, "  </subkey>\n");
     }
     _gpgme_data_append_string ( d, "</GnupgKeyblock>\n" );
@@ -869,6 +905,12 @@ gpgme_key_get_ulong_attr ( GpgmeKey key, GpgmeAttr what,
         if (k) 
             val = k->timestamp < 0? 0L:(unsigned long)k->timestamp;
         break;
+      case GPGME_ATTR_EXPIRE: 
+        for (k=&key->keys; k && idx; k=k->next, idx-- )
+            ;
+        if (k) 
+            val = k->expires_at < 0? 0L:(unsigned long)k->expires_at;
+        break;
       case GPGME_ATTR_VALIDITY:
         for (u=key->uids; u && idx; u=u->next, idx-- )
             ;
