@@ -42,7 +42,7 @@
 void
 _gpgme_sema_subsystem_init ()
 {
-#warning Posix semaphore support has not yet been implemented 
+    /* FIXME: Posix semaphore support has not yet been implemented */
 }
 
 
