@@ -1,6 +1,6 @@
 /* rungpg.c 
  *	Copyright (C) 2000 Werner Koch (dd9jn)
- *      Copyright (C) 2001 g10 Code GmbH
+ *      Copyright (C) 2001, 2002 g10 Code GmbH
  *
  * This file is part of GPGME.
  *
@@ -118,23 +118,9 @@ struct gpg_object_s {
     } cmd;
 };
 
-struct reap_s {
-    struct reap_s *next;
-    int pid;
-    time_t entered;
-    int term_send;
-};
-
-static struct reap_s *reap_list;
-DEFINE_STATIC_LOCK (reap_list_lock);
-
-
 static void free_argv ( char **argv );
 static void free_fd_data_map ( struct fd_data_map_s *fd_data_map );
 
-static int gpg_inbound_handler ( void *opaque, int pid, int fd );
-static int gpg_outbound_handler ( void *opaque, int pid, int fd );
-
 static int gpg_status_handler ( void *opaque, int pid, int fd );
 static GpgmeError read_status ( GpgObject gpg );
 
@@ -177,8 +163,25 @@ close_notify_handler ( int fd, void *opaque )
     }
 }
 
+const char *
+_gpgme_gpg_get_version (void)
+{
+  static const char *gpg_version;
 
+  /* FIXME: Locking.  */
+  if (!gpg_version)
+    gpg_version = _gpgme_get_program_version (_gpgme_get_gpg_path ());
 
+  return gpg_version;
+}
+
+GpgmeError
+_gpgme_gpg_check_version (void)
+{
+  return _gpgme_compare_versions (_gpgme_gpg_get_version (),
+                                  NEED_GPG_VERSION)
+    ? 0 : mk_error (Invalid_Engine);
+}
 
 GpgmeError
 _gpgme_gpg_new ( GpgObject *r_gpg )
@@ -244,102 +247,41 @@ _gpgme_gpg_new ( GpgObject *r_gpg )
 
 
 void
-_gpgme_gpg_release ( GpgObject gpg )
+_gpgme_gpg_release (GpgObject gpg)
 {
-    if ( !gpg )
-        return;
-    xfree (gpg->status.buffer);
-    xfree (gpg->colon.buffer);
-    if ( gpg->argv )
-        free_argv (gpg->argv);
-    xfree (gpg->cmd.keyword);
+  if (!gpg)
+    return;
 
-    if (gpg->pid != -1) 
-        _gpgme_remove_proc_from_wait_queue ( gpg->pid );
-    if (gpg->status.fd[0] != -1 )
-        _gpgme_io_close (gpg->status.fd[0]);
-    if (gpg->status.fd[1] != -1 )
-        _gpgme_io_close (gpg->status.fd[1]);
-    if (gpg->colon.fd[0] != -1 )
-        _gpgme_io_close (gpg->colon.fd[0]);
-    if (gpg->colon.fd[1] != -1 )
-        _gpgme_io_close (gpg->colon.fd[1]);
-    free_fd_data_map (gpg->fd_data_map);
-    if (gpg->running) {
-        int pid = gpg->pid;
-        struct reap_s *r;
+  while (gpg->arglist)
+    {
+      struct arg_and_data_s *next = gpg->arglist->next;
 
-        /* resuse the memory, so that we don't need to allocate another
-         * mem block and have to handle errors */
-        assert (sizeof *r < sizeof *gpg );
-        r = (void*)gpg;
-        memset (r, 0, sizeof *r);
-        r->pid = pid;
-        r->entered = time (NULL);
-        LOCK(reap_list_lock);
-        r->next = reap_list;
-        reap_list = r;
-        UNLOCK(reap_list_lock);
+      xfree (gpg->arglist);
+      gpg->arglist = next;
     }
-    else
-        xfree (gpg);
-}
 
+  xfree (gpg->status.buffer);
+  xfree (gpg->colon.buffer);
+  if (gpg->argv)
+    free_argv (gpg->argv);
+  xfree (gpg->cmd.cb_data);
+  xfree (gpg->cmd.keyword);
 
-static void
-do_reaping (void)
-{
-    struct reap_s *r, *rlast;
-    static time_t last_check;
-    time_t cur_time = time (NULL);
-
-    /* a race does not matter here */
-    if (!last_check)
-        last_check = time(NULL);
-
-    if (last_check >= cur_time)
-        return;  /* we check only every second */
-
-    /* fixme: it would be nice if to have a TRYLOCK here */
-    LOCK (reap_list_lock);  
-    for (r=reap_list,rlast=NULL; r ; rlast=r, r=r?r->next:NULL) {
-        int dummy1, dummy2;
-
-        if ( _gpgme_io_waitpid (r->pid, 0, &dummy1, &dummy2) ) {
-            /* process has terminated - remove it from the queue */
-            void *p = r;
-            if (!rlast) {
-                reap_list = r->next;
-                r = reap_list;
-            }
-            else {
-                rlast->next = r->next;
-                r = rlast;
-            }
-            xfree (p);
-        }
-        else if ( !r->term_send ) {
-            if( r->entered+1 >= cur_time ) {
-                _gpgme_io_kill ( r->pid, 0);
-                r->term_send = 1;
-                r->entered = cur_time;
-            }
-        }
-        else {
-            /* give it 5 second before we are going to send the killer */
-            if ( r->entered+5 >= cur_time ) {
-                _gpgme_io_kill (r->pid, 1);
-                r->entered = cur_time; /* just in case we have to repat it */
-            }
-        }
-    }
-    UNLOCK (reap_list_lock);  
-}
-
-void
-_gpgme_gpg_housecleaning ()
-{
-    do_reaping ();
+  if (gpg->pid != -1) 
+    _gpgme_remove_proc_from_wait_queue (gpg->pid);
+  if (gpg->status.fd[0] != -1)
+    _gpgme_io_close (gpg->status.fd[0]);
+  if (gpg->status.fd[1] != -1)
+    _gpgme_io_close (gpg->status.fd[1]);
+  if (gpg->colon.fd[0] != -1)
+    _gpgme_io_close (gpg->colon.fd[0]);
+  if (gpg->colon.fd[1] != -1)
+    _gpgme_io_close (gpg->colon.fd[1]);
+  free_fd_data_map (gpg->fd_data_map);
+  if (gpg->running)
+    _gpgme_engine_add_child_to_reap_list (gpg, sizeof *gpg, gpg->pid);
+  else
+    xfree (gpg);
 }
 
 void
@@ -775,8 +717,11 @@ _gpgme_gpg_spawn( GpgObject gpg, void *opaque )
     int pid;
     struct spawn_fd_item_s *fd_child_list, *fd_parent_list;
 
-    if ( !gpg )
-        return mk_error (Invalid_Value);
+    if (!gpg)
+      return mk_error (Invalid_Value);
+
+    if (! _gpgme_get_gpg_path ())
+      return mk_error (Invalid_Engine);
 
     /* Kludge, so that we don't need to check the return code of
      * all the gpgme_gpg_add_arg().  we bail out here instead */
@@ -790,9 +735,9 @@ _gpgme_gpg_spawn( GpgObject gpg, void *opaque )
     if ( rc )
         return rc;
 
-    n = 4; /* status fd, 2*colon_fd and end of list */
+    n = 3; /* status_fd, colon_fd and end of list */
     for (i=0; gpg->fd_data_map[i].data; i++ ) 
-        n += 2;
+        n++;
     fd_child_list = xtrycalloc ( n+n, sizeof *fd_child_list );
     if (!fd_child_list)
         return mk_error (Out_Of_Core);
@@ -800,21 +745,12 @@ _gpgme_gpg_spawn( GpgObject gpg, void *opaque )
 
     /* build the fd list for the child */
     n=0;
-    fd_child_list[n].fd = gpg->status.fd[0]; 
-    fd_child_list[n].dup_to = -1;
-    n++;
     if ( gpg->colon.fnc ) {
-        fd_child_list[n].fd = gpg->colon.fd[0];
-        fd_child_list[n].dup_to = -1;
-        n++;
         fd_child_list[n].fd = gpg->colon.fd[1]; 
         fd_child_list[n].dup_to = 1; /* dup to stdout */
         n++;
     }
     for (i=0; gpg->fd_data_map[i].data; i++ ) {
-        fd_child_list[n].fd = gpg->fd_data_map[i].fd;
-        fd_child_list[n].dup_to = -1;
-        n++;
         if (gpg->fd_data_map[i].dup_to != -1) {
             fd_child_list[n].fd = gpg->fd_data_map[i].peer_fd;
             fd_child_list[n].dup_to = gpg->fd_data_map[i].dup_to;
@@ -888,7 +824,7 @@ _gpgme_gpg_spawn( GpgObject gpg, void *opaque )
         if ( _gpgme_register_pipe_handler (
                  opaque, 
                  gpg->fd_data_map[i].inbound?
-                       gpg_inbound_handler:gpg_outbound_handler,
+		 _gpgme_data_inbound_handler:_gpgme_data_outbound_handler,
                  gpg->fd_data_map[i].data,
                  pid, gpg->fd_data_map[i].fd,
                  gpg->fd_data_map[i].inbound )
@@ -909,148 +845,23 @@ _gpgme_gpg_spawn( GpgObject gpg, void *opaque )
 
 
 static int
-gpg_inbound_handler ( void *opaque, int pid, int fd )
+gpg_status_handler (void *opaque, int pid, int fd)
 {
-    GpgmeData dh = opaque;
-    GpgmeError err;
-    int nread;
-    char buf[200];
+  GpgObject gpg = opaque;
+  int err;
 
-    assert ( _gpgme_data_get_mode (dh) == GPGME_DATA_MODE_IN );
-
-    nread = _gpgme_io_read (fd, buf, 200 );
-    if ( nread < 0 ) {
-        DEBUG3 ("read_mem_data: read failed on fd %d (n=%d): %s",
-                 fd, nread, strerror (errno) );
-        return 1;
+  assert (fd == gpg->status.fd[0]);
+  err = read_status (gpg);
+  if (err)
+    {
+      /* XXX Horrible kludge.  We really must not make use of
+	 fnc_value.  */
+      GpgmeCtx ctx = (GpgmeCtx) gpg->status.fnc_value;
+      ctx->error = err;
+      DEBUG1 ("gpg_handler: read_status problem %d\n - stop", err);
+      return 1;
     }
-    else if (!nread)
-        return 1; /* eof */
-
-    /* We could improve this with a GpgmeData function which takes
-     * the read function or provides a memory area for writing to it.
-     */
-    
-    err = _gpgme_data_append ( dh, buf, nread );
-    if ( err ) {
-        DEBUG1 ("_gpgme_append_data failed: %s\n",
-                 gpgme_strerror(err));
-        /* Fixme: we should close the pipe or read it to /dev/null in
-         * this case. Returnin EOF is not sufficient */
-        return 1;
-    }
-
-    return 0;
-}
-
-
-static int
-write_mem_data ( GpgmeData dh, int fd )
-{
-    size_t nbytes;
-    int  nwritten; 
-
-    nbytes = dh->len - dh->readpos;
-    if ( !nbytes ) {
-        _gpgme_io_close (fd);
-        return 1;
-    }
-    
-    /* FIXME: Arggg, the pipe blocks on large write request, although
-     * select told us that it is okay to write - need to figure out
-     * why this happens?  Stevens says nothing about this problem (or
-     * is it my Linux kernel 2.4.0test1)
-     * To avoid that we have set the pipe to nonblocking.
-     */
-
-    nwritten = _gpgme_io_write ( fd, dh->data+dh->readpos, nbytes );
-    if (nwritten == -1 && errno == EAGAIN )
-        return 0;
-    if ( nwritten < 1 ) {
-        DEBUG3 ("write_mem_data(%d): write failed (n=%d): %s",
-                fd, nwritten, strerror (errno) );
-        _gpgme_io_close (fd);
-        return 1;
-    }
-
-    dh->readpos += nwritten;
-    return 0;
-}
-
-static int
-write_cb_data ( GpgmeData dh, int fd )
-{
-    size_t nbytes;
-    int  err, nwritten; 
-    char buffer[512];
-
-    err = gpgme_data_read ( dh, buffer, DIM(buffer), &nbytes );
-    if (err == GPGME_EOF) {
-        _gpgme_io_close (fd);
-        return 1;
-    }
-    
-    nwritten = _gpgme_io_write ( fd, buffer, nbytes );
-    if (nwritten == -1 && errno == EAGAIN )
-        return 0;
-    if ( nwritten < 1 ) {
-        DEBUG3 ("write_cb_data(%d): write failed (n=%d): %s",
-                fd, nwritten, strerror (errno) );
-        _gpgme_io_close (fd);
-        return 1;
-    }
-
-    if ( nwritten < nbytes ) {
-        /* ugly, ugly: It does currently only for for MEM type data */
-        if ( _gpgme_data_unread (dh, buffer + nwritten, nbytes - nwritten ) )
-            DEBUG1 ("wite_cb_data: unread of %d bytes failed\n",
-                     nbytes - nwritten );
-        _gpgme_io_close (fd);
-        return 1;
-    }
-
-    return 0;
-}
-
-
-static int
-gpg_outbound_handler ( void *opaque, int pid, int fd )
-{
-    GpgmeData dh = opaque;
-
-    assert ( _gpgme_data_get_mode (dh) == GPGME_DATA_MODE_OUT );
-    switch ( gpgme_data_get_type (dh) ) {
-      case GPGME_DATA_TYPE_MEM:
-        if ( write_mem_data ( dh, fd ) )
-            return 1; /* ready */
-        break;
-      case GPGME_DATA_TYPE_CB:
-        if (write_cb_data (dh, fd))
-            return 1; /* ready */
-        break;
-      default:
-        assert (0);
-    }
-
-    return 0;
-}
-
-
-
-static int
-gpg_status_handler ( void *opaque, int pid, int fd )
-{
-    GpgObject gpg = opaque;
-    int rc = 0;
-
-    assert ( fd == gpg->status.fd[0] );
-    rc = read_status ( gpg );
-    if ( rc ) {
-        DEBUG1 ("gpg_handler: read_status problem %d\n - stop", rc);
-        return 1;
-    }
-
-    return gpg->status.eof;
+  return gpg->status.eof;
 }
 
 
@@ -1275,7 +1086,7 @@ static GpgmeError
 pipemode_copy (char *buffer, size_t length, size_t *nread, GpgmeData data )
 {
     GpgmeError err;
-    int nbytes;
+    size_t nbytes;
     char tmp[1000], *s, *d;
 
     /* we can optimize this whole thing but for now we just
@@ -1403,6 +1214,384 @@ command_cb ( void *opaque, char *buffer, size_t length, size_t *nread )
     return 0;
 }
 
+GpgmeError
+_gpgme_gpg_op_decrypt (GpgObject gpg, GpgmeData ciph, GpgmeData plain)
+{
+  GpgmeError err;
+
+  err = _gpgme_gpg_add_arg (gpg, "--decrypt");
+
+  /* Tell the gpg object about the data.  */
+  if (!err)
+    err = _gpgme_gpg_add_arg (gpg, "--output");
+  if (!err)
+    err = _gpgme_gpg_add_arg (gpg, "-");
+  if (!err)
+    err = _gpgme_gpg_add_data (gpg, plain, 1);
+  if (!err)
+    err = _gpgme_gpg_add_data (gpg, ciph, 0);
+
+  return err;
+}
+
+GpgmeError
+_gpgme_gpg_op_delete (GpgObject gpg, GpgmeKey key, int allow_secret)
+{
+  GpgmeError err;
+
+  err = _gpgme_gpg_add_arg (gpg, allow_secret
+			    ? "--delete-secret-and-public-key"
+			    : "--delete-key");
+  if (!err)
+    err = _gpgme_gpg_add_arg (gpg, "--");
+  if (!err)
+    {
+      const char *s = gpgme_key_get_string_attr (key, GPGME_ATTR_FPR, NULL, 0);
+      if (!s)
+	err = mk_error (Invalid_Key);
+      else
+	err = _gpgme_gpg_add_arg (gpg, s);
+    }
+
+  return err;
+}
 
 
+static GpgmeError
+_gpgme_append_gpg_args_from_recipients (GpgObject gpg,
+					const GpgmeRecipients rset)
+{
+  GpgmeError err = 0;
+  struct user_id_s *r;
 
+  assert (rset);
+  for (r = rset->list; r; r = r->next)
+    {
+      err = _gpgme_gpg_add_arg (gpg, "-r");
+      if (!err)
+	_gpgme_gpg_add_arg (gpg, r->name);
+      if (err)
+	break;
+    }    
+  return err;
+}
+
+
+static GpgmeError
+_gpgme_append_gpg_args_from_signers (GpgObject gpg,
+				     GpgmeCtx ctx /* FIXME */)
+{
+  GpgmeError err = 0;
+  int i;
+  GpgmeKey key;
+
+  for (i = 0; (key = gpgme_signers_enum (ctx, i)); i++)
+    {
+      const char *s = gpgme_key_get_string_attr (key, GPGME_ATTR_KEYID,
+						 NULL, 0);
+      if (s)
+	{
+	  if (!err)
+	    err = _gpgme_gpg_add_arg (gpg, "-u");
+	  if (!err)
+	    err = _gpgme_gpg_add_arg (gpg, s);
+	}
+      gpgme_key_unref (key);
+      if (err) break;
+    }
+  return err;
+}
+
+
+GpgmeError
+_gpgme_gpg_op_encrypt (GpgObject gpg, GpgmeRecipients recp,
+		       GpgmeData plain, GpgmeData ciph, int use_armor)
+{
+  GpgmeError err;
+
+  err = _gpgme_gpg_add_arg (gpg, "--encrypt");
+  if (!err && use_armor)
+    err = _gpgme_gpg_add_arg (gpg, "--armor");
+
+  /* If we know that all recipients are valid (full or ultimate trust)
+   * we can suppress further checks */
+  if (!err && _gpgme_recipients_all_valid (recp))
+    err = _gpgme_gpg_add_arg (gpg, "--always-trust");
+
+  if (!err)
+    err = _gpgme_append_gpg_args_from_recipients (gpg, recp);
+
+  /* Tell the gpg object about the data.  */
+  if (!err)
+    err = _gpgme_gpg_add_arg (gpg, "--output");
+  if (!err)
+    err = _gpgme_gpg_add_arg (gpg, "-");
+  if (!err)
+    err = _gpgme_gpg_add_data (gpg, ciph, 1);
+  if (!err)
+    err = _gpgme_gpg_add_arg (gpg, "--");
+  if (!err)
+    err = _gpgme_gpg_add_data (gpg, plain, 0);
+
+  return err;
+}
+
+GpgmeError
+_gpgme_gpg_op_encrypt_sign (GpgObject gpg, GpgmeRecipients recp,
+			    GpgmeData plain, GpgmeData ciph, int use_armor,
+			    GpgmeCtx ctx /* FIXME */)
+{
+  GpgmeError err;
+
+  err = _gpgme_gpg_add_arg (gpg, "--encrypt");
+  if (!err)
+    err = _gpgme_gpg_add_arg (gpg, "--sign");
+  if (!err && use_armor)
+    err = _gpgme_gpg_add_arg (gpg, "--armor");
+
+  /* If we know that all recipients are valid (full or ultimate trust)
+   * we can suppress further checks */
+  if (!err && _gpgme_recipients_all_valid (recp))
+    err = _gpgme_gpg_add_arg (gpg, "--always-trust");
+
+  if (!err)
+    err = _gpgme_append_gpg_args_from_recipients (gpg, recp);
+
+  if (!err)
+    err = _gpgme_append_gpg_args_from_signers (gpg, ctx);
+
+  /* Tell the gpg object about the data.  */
+  if (!err)
+    err = _gpgme_gpg_add_arg (gpg, "--output");
+  if (!err)
+    err = _gpgme_gpg_add_arg (gpg, "-");
+  if (!err)
+    err = _gpgme_gpg_add_data (gpg, ciph, 1);
+  if (!err)
+    err = _gpgme_gpg_add_arg (gpg, "--");
+  if (!err)
+    err = _gpgme_gpg_add_data (gpg, plain, 0);
+
+  return err;
+}
+
+GpgmeError
+_gpgme_gpg_op_export (GpgObject gpg, GpgmeRecipients recp,
+		      GpgmeData keydata, int use_armor)
+{
+  GpgmeError err;
+
+  err = _gpgme_gpg_add_arg (gpg, "--export");
+  if (!err && use_armor)
+    err = _gpgme_gpg_add_arg (gpg, "--armor");
+  if (!err)
+    err = _gpgme_gpg_add_data (gpg, keydata, 1);
+  if (!err)
+    err = _gpgme_gpg_add_arg (gpg, "--");
+
+  if (!err)
+    {
+      void *ec;
+      const char *s;
+
+      err = gpgme_recipients_enum_open (recp, &ec);
+      while (!err && (s = gpgme_recipients_enum_read (recp, &ec)))
+	err = _gpgme_gpg_add_arg (gpg, s);
+      if (!err)
+	err = gpgme_recipients_enum_close (recp, &ec);
+    }
+
+  return err;
+}
+
+GpgmeError
+_gpgme_gpg_op_genkey (GpgObject gpg, GpgmeData help_data, int use_armor,
+		      GpgmeData pubkey, GpgmeData seckey)
+{
+  GpgmeError err;
+
+  if (!gpg)
+    return mk_error (Invalid_Value);
+
+  /* We need a special mechanism to get the fd of a pipe here, so
+   * that we can use this for the %pubring and %secring parameters.
+   * We don't have this yet, so we implement only the adding to the
+   * standard keyrings */
+  if (pubkey || seckey)
+    return err = mk_error (Not_Implemented);
+
+  err = _gpgme_gpg_add_arg (gpg, "--gen-key");
+  if (!err && use_armor)
+    err = _gpgme_gpg_add_arg (gpg, "--armor");
+  if (!err)
+    err = _gpgme_gpg_add_data (gpg, help_data, 0);
+
+  return err;
+}
+
+GpgmeError
+_gpgme_gpg_op_import (GpgObject gpg, GpgmeData keydata)
+{
+  GpgmeError err;
+
+  err = _gpgme_gpg_add_arg (gpg, "--import");
+  if (!err)
+    err = _gpgme_gpg_add_data (gpg, keydata, 0);
+
+  return err;
+}
+
+
+GpgmeError
+_gpgme_gpg_op_keylist (GpgObject gpg, const char *pattern, int secret_only,
+		       int keylist_mode)
+{
+  GpgmeError err;
+
+  err = _gpgme_gpg_add_arg (gpg, "--with-colons");
+  if (!err)
+    err = _gpgme_gpg_add_arg (gpg, "--fixed-list-mode");
+  if (!err)
+    err = _gpgme_gpg_add_arg (gpg, "--with-fingerprint");
+  if (!err)
+    err = _gpgme_gpg_add_arg (gpg, secret_only ? "--list-secret-keys"
+			      : "--list-keys");
+  
+  /* Tell the gpg object about the data */
+  if (!err)
+    err = _gpgme_gpg_add_arg (gpg, "--");
+  if (!err && pattern && *pattern)
+    err = _gpgme_gpg_add_arg (gpg, pattern);
+
+  return err;
+}
+
+
+GpgmeError
+_gpgme_gpg_op_keylist_ext (GpgObject gpg, const char *pattern[],
+			   int secret_only, int reserved, int keylist_mode)
+{
+  GpgmeError err;
+
+  if (reserved)
+    return mk_error (Invalid_Value);
+
+  err = _gpgme_gpg_add_arg (gpg, "--with-colons");
+  if (!err)
+    err = _gpgme_gpg_add_arg (gpg, "--fixed-list-mode");
+  if (!err)
+    err = _gpgme_gpg_add_arg (gpg, "--with-fingerprint");
+  if (!err)
+    err = _gpgme_gpg_add_arg (gpg, secret_only ? "--list-secret-keys"
+			      : "--list-keys");
+  
+  /* Tell the gpg object about the data */
+  if (!err)
+    err = _gpgme_gpg_add_arg (gpg, "--");
+  if (!err && pattern && *pattern)
+    {
+      while (*pattern)
+	err = _gpgme_gpg_add_arg (gpg, *(pattern++));
+    }
+
+  return err;
+}
+
+
+GpgmeError
+_gpgme_gpg_op_sign (GpgObject gpg, GpgmeData in, GpgmeData out,
+		    GpgmeSigMode mode, int use_armor,
+		    int use_textmode, GpgmeCtx ctx /* FIXME */)
+{
+  GpgmeError err;
+
+  if (mode == GPGME_SIG_MODE_CLEAR)
+    err = _gpgme_gpg_add_arg (gpg, "--clearsign");
+  else
+    {
+      err = _gpgme_gpg_add_arg (gpg, "--sign");
+      if (!err && mode == GPGME_SIG_MODE_DETACH)
+	err = _gpgme_gpg_add_arg (gpg, "--detach");
+      if (!err && use_armor)
+	err = _gpgme_gpg_add_arg (gpg, "--armor");
+      if (!err && use_textmode)
+	_gpgme_gpg_add_arg (gpg, "--textmode");
+    }
+
+  if (!err)
+    err = _gpgme_append_gpg_args_from_signers (gpg, ctx);
+
+  /* Tell the gpg object about the data.  */
+  if (!err)
+    err = _gpgme_gpg_add_data (gpg, in, 0);
+  if (!err)
+    err = _gpgme_gpg_add_data (gpg, out, 1);
+
+  return err;
+}
+
+GpgmeError
+_gpgme_gpg_op_trustlist (GpgObject gpg, const char *pattern)
+{
+  GpgmeError err;
+
+  err = _gpgme_gpg_add_arg (gpg, "--with-colons");
+  if (!err)
+    err = _gpgme_gpg_add_arg (gpg, "--list-trust-path");
+  
+  /* Tell the gpg object about the data */
+  if (!err)
+    err = _gpgme_gpg_add_arg (gpg, "--");
+  if (!err)
+    err = _gpgme_gpg_add_arg (gpg, pattern);
+
+  return err;
+}
+
+GpgmeError
+_gpgme_gpg_op_verify (GpgObject gpg, GpgmeData sig, GpgmeData text)
+{
+  GpgmeError err = 0;
+
+  if (_gpgme_data_get_mode (text) == GPGME_DATA_MODE_IN)
+    {
+      /* Normal or cleartext signature.  */
+
+      err = _gpgme_gpg_add_arg (gpg, "--output");
+      if (!err)
+	err = _gpgme_gpg_add_arg (gpg, "-");
+      if (!err)
+	err = _gpgme_gpg_add_data (gpg, sig, 0);
+      if (!err)
+	err = _gpgme_gpg_add_data (gpg, text, 1);
+    }
+  else
+    {
+      if (gpg->pm.used)
+	{
+	  err = _gpgme_gpg_add_arg (gpg, gpg->pm.used ? "--pipemode" : "--verify");
+	  if (!err)
+	    err = _gpgme_gpg_add_arg (gpg, "--");
+	  if (!err)
+	    err = _gpgme_gpg_add_pm_data (gpg, sig, 0);
+	  if (!err)
+	    err = _gpgme_gpg_add_pm_data (gpg, text, 1);
+	}
+      else
+	{
+	  err = _gpgme_gpg_add_arg (gpg, "--verify");
+	  if (!err)
+	    err = _gpgme_gpg_add_arg (gpg, "--");
+	  if (!err)
+	    err = _gpgme_gpg_add_data (gpg, sig, -1);
+	  if (text)
+	    {
+	      if (!err)
+		err = _gpgme_gpg_add_arg (gpg, "-");
+	      if (!err)
+		err = _gpgme_gpg_add_data (gpg, text, 0);
+	    }
+	}
+    }
+  return err;
+}
